// -----------------------------------------------------------------------------
// |    ___  ___  _  _ _     ___                                        _      |
// |   / __>| . \| || \ |   | __>_ _  ___ ._ _ _  ___  _ _ _  ___  _ _ | |__   |
// |   \__ \|  _/| ||   |   | _>| '_><_> || ' ' |/ ._>| | | |/ . \| '_>| / /   |
// |   <___/|_|  |_||_\_|   |_| |_|  <___||_|_|_|\___.|__/_/ \___/|_|  |_\_\   |
// |                                                                           |
// |---------------------------------------------------------------------------|
//
// http://spinframework.sourceforge.net
// Copyright (C) 2009 Mike Wozniewski, Zack Settel
//
// Developed/Maintained by:
//    Mike Wozniewski (http://www.mikewoz.com)
//    Zack Settel (http://www.sheefa.net/zack)
//
// Principle Partners:
//    Shared Reality Lab, McGill University (http://www.cim.mcgill.ca/sre)
//    La Societe des Arts Technologiques (http://www.sat.qc.ca)
//
// Funding by:
//    NSERC/Canada Council for the Arts - New Media Initiative
//    Heritage Canada
//    Ministere du Developpement economique, de l'Innovation et de l'Exportation
//
// -----------------------------------------------------------------------------
//  This file is part of the SPIN Framework.
//
//  SPIN Framework is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  SPIN Framework is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with SPIN Framework. If not, see <http://www.gnu.org/licenses/>.
// -----------------------------------------------------------------------------

#ifndef SwitchNode_H_
#define SwitchNode_H_

#include "groupnode.h"

namespace osg {
    class Switch;
}

namespace spin
{

/**
 * \brief Allows for selectively enabling/disabling child nodes
 *
 */
class SwitchNode : public GroupNode
{

public:

    SwitchNode(SceneManager *sceneManager, const char* initID);
    virtual ~SwitchNode();

    /**
     * IMPORTANT:
     * subclasses of ReferencedNode are allowed to contain complicated subgraphs,
     * and can also change their attachmentNode so that children are attached
     * anywhere in that subgraph. If that is the case, the updateNodePath()
     * function MUST be overridden, and extra nodes must be manually pushed onto
     * currentNodePath_.
     */
    virtual void updateNodePath();

    /**
     * This enables/disables a child of a certain id
     */
    void setEnabled (const char* id, int enabled);
    
    /**
     * Check if a particular node is a child and is enabled
     */
    bool isEnabled (ReferencedNode* n);

    /**
     * Set all child noded to be either disabled or enabled
     */
    void setAll(int enabled);

    /**
     * For each subclass of ReferencedNode, we override the getState() method to
     * fill the vector with the correct set of methods for this particular node
     */
    virtual std::vector<lo_message> getState() const;

private:

    osg::ref_ptr<osg::Switch> switcher;
};


} // end of namespace spin


#endif
